{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "9c0ffcb5",
   "metadata": {},
   "source": [
    "# Singleton\n",
    "\n",
    "[![Twitter Handle](https://img.shields.io/badge/Twitter-@gaohongnan-blue?style=social&logo=twitter)](https://twitter.com/gaohongnan)\n",
    "[![LinkedIn Profile](https://img.shields.io/badge/@gaohongnan-blue?style=social&logo=linkedin)](https://linkedin.com/in/gao-hongnan)\n",
    "[![GitHub Profile](https://img.shields.io/badge/GitHub-gao--hongnan-lightgrey?style=social&logo=github)](https://github.com/gao-hongnan)\n",
    "![Tag](https://img.shields.io/badge/Tag-Organized_Chaos-orange)\n",
    "\n",
    "```{contents}\n",
    ":local:\n",
    "```\n",
    "\n",
    "````{tab} **logger.py**\n",
    "```python\n",
    "from __future__ import annotations\n",
    "\n",
    "import threading\n",
    "from typing import Type\n",
    "\n",
    "\n",
    "class Logger:\n",
    "    _instance: Logger | None = None\n",
    "    _lock: threading.Lock = threading.Lock()\n",
    "\n",
    "    def __new__(cls: Type[Logger]) -> Logger:  # noqa: PYI034\n",
    "        if cls._instance is None:\n",
    "            with cls._lock:\n",
    "                if cls._instance is None:\n",
    "                    cls._instance = object.__new__(cls)\n",
    "        return cls._instance\n",
    "\n",
    "    def __init__(self) -> None:\n",
    "        \"\"\"The initialized flag is used to prevent the __init__ method from\n",
    "        being called more than once.\n",
    "        \"\"\"\n",
    "        if not hasattr(self, \"initialized\"):\n",
    "            self.initialized = True\n",
    "            self.log(f\"{self.__class__.__name__} initialized with id={id(self)}\")\n",
    "\n",
    "    def __eq__(self, other: object) -> bool:\n",
    "        if not isinstance(other, Logger):\n",
    "            return NotImplemented\n",
    "        return id(self) == id(other)\n",
    "\n",
    "    def log(self, message: str) -> None:\n",
    "        print(f\"LOG: {message}\")\n",
    "```\n",
    "````\n",
    "\n",
    "````{tab} **core.py**\n",
    "```python\n",
    "from omnixamples.software_engineering.design_patterns.singleton.logger import Logger\n",
    "\n",
    "logger1 = Logger()\n",
    "logger2 = Logger()\n",
    "\n",
    "assert logger1 is logger2\n",
    "assert logger1 == logger2\n",
    "print(f\"logger1 id={id(logger1)} | logger2 id={id(logger2)}\")\n",
    "```\n",
    "````\n",
    "\n",
    "## Project Structure Overview\n",
    "\n",
    "To contextualize the implementation, let's visualize the directory structure of\n",
    "the project:\n",
    "\n",
    "```text\n",
    "omnixamples/\n",
    "└── software_engineering/\n",
    "    └── design_patterns/\n",
    "        └── singleton/\n",
    "            ├── __init__.py\n",
    "            ├── logger.py\n",
    "            └── core.py\n",
    "```\n",
    "\n",
    "-   **`omnixamples/`**: Root directory of the project.\n",
    "-   **`logger.py`**: Contains the `Logger` Singleton class implementation.\n",
    "-   **`core.py`**: Demonstrates how to use the `Logger` Singleton.\n",
    "\n",
    "## What is the Singleton Pattern?\n",
    "\n",
    "The **Singleton** is a\n",
    "[**creational design pattern**](https://refactoring.guru/design-patterns/singleton)\n",
    "that ensures a class has **only one instance** and provides a **global point of\n",
    "access** to that instance. This is particularly useful when exactly one object\n",
    "is needed to coordinate actions across the system, such as:\n",
    "\n",
    "-   **Logging Systems**\n",
    "-   **Configuration Managers**\n",
    "\n",
    "## Why Use a Singleton?\n",
    "\n",
    "Consider a global settings object `Settings` hosting your platform definitions,\n",
    "and you really want to ensure a single source of truth here. Same idea goes for\n",
    "logger, so technically you can also instantiate the logger with the desired\n",
    "configurations inside, say `__init__.py` and then use it anywhere in the\n",
    "project.\n",
    "\n",
    "## The Singleton Pattern in Detail\n",
    "\n",
    "### Class Variables\n",
    "\n",
    "The key to the singleton pattern is the use of class variables because they are\n",
    "shared across all instances of the class - like a global variable.\n",
    "\n",
    "```python\n",
    "_instance: Logger | None = None\n",
    "_lock: threading.Lock = threading.Lock()\n",
    "```\n",
    "\n",
    "-   **`_instance: Logger | None = None`:**\n",
    "\n",
    "    -   **Purpose:** Holds the singleton instance of the `Logger` class.\n",
    "    -   **Type Annotation:** Indicates that `_instance` can be either a `Logger`\n",
    "        instance or `None`.\n",
    "    -   **Initial Value:** Set to `None`, meaning no instance exists at the\n",
    "        start.\n",
    "\n",
    "-   **`_lock: threading.Lock = threading.Lock()`:**\n",
    "    -   **Purpose:** A thread lock to ensure that only one thread can create an\n",
    "        instance at a time.\n",
    "    -   **Type Annotation:** Specifies that `_lock` is of type `threading.Lock`.\n",
    "    -   **Initialization:** Creates a new lock instance.\n",
    "\n",
    "### Overriding the `__new__` Method\n",
    "\n",
    "```python\n",
    "def __new__(cls: Type[Logger]) -> Logger:  # noqa: PYI034\n",
    "    if cls._instance is None:\n",
    "        with cls._lock:\n",
    "            if cls._instance is None:\n",
    "                cls._instance = object.__new__(cls)\n",
    "    return cls._instance\n",
    "```\n",
    "\n",
    "-   **Purpose of `__new__`:**\n",
    "\n",
    "    -   `__new__` is a **special method** responsible for **creating a new\n",
    "        instance** of a class.\n",
    "    -   It is called **before** `__init__`.\n",
    "    -   Suitable for implementing singletons since it controls the instantiation\n",
    "        process.\n",
    "\n",
    "-   **Implementation Details:**\n",
    "\n",
    "    1. **Check for Existing Instance:**\n",
    "        - `if cls._instance is None:`: Determines if an instance already exists.\n",
    "    2. **Acquire Lock for Thread Safety:**\n",
    "        - `with cls._lock:`: Ensures that only one thread can execute the block\n",
    "          at a time.\n",
    "    3. **Double-Checked Locking:**\n",
    "        - Inside the locked block, it **rechecks** `if cls._instance is None:`\n",
    "          to prevent race conditions where multiple threads might have passed\n",
    "          the first check simultaneously.\n",
    "    4. **Create New Instance:**\n",
    "        - `cls._instance = object.__new__(cls)`: Calls the base class (`object`)\n",
    "          `__new__` method to create a new instance and assigns it to\n",
    "          `_instance`.\n",
    "    5. **Return the Instance:**\n",
    "        - Regardless of whether a new instance was created or an existing one is\n",
    "          used, `cls._instance` is returned.\n",
    "\n",
    "### **e. Overriding the `__init__` Method**\n",
    "\n",
    "```python\n",
    "def __init__(self) -> None:\n",
    "    \"\"\"The initialized flag is used to prevent the __init__ method from\n",
    "    being called more than once.\n",
    "    \"\"\"\n",
    "    if not hasattr(self, \"initialized\"):\n",
    "        self.initialized = True\n",
    "        self.log(f\"{self.__class__.__name__} initialized with id={id(self)}\")\n",
    "```\n",
    "\n",
    "-   **Purpose of `__init__`:**\n",
    "\n",
    "    -   `__init__` initializes the instance after it has been created by\n",
    "        `__new__`.\n",
    "    -   In the singleton pattern, it's crucial to prevent re-initialization if\n",
    "        the instance already exists. Why so? Because the `__init__` method is\n",
    "        called every time an instance is created, and we only want to run it\n",
    "        once at some cases to prevent mutation of instance variables.\n",
    "\n",
    "## Running Core\n",
    "\n",
    "```python\n",
    "LOG: Logger initialized with id=4379880944\n",
    "LOG: Logger __init__ called again with id=4379880944\n",
    "logger1 id=4379880944 | logger2 id=4379880944\n",
    "```\n",
    "\n",
    "We see that the memory address of the two logger instances are the same, which\n",
    "indicates that they are the same instance - hence the singleton pattern is\n",
    "working as expected.\n",
    "\n",
    "## Thread Safety Considerations\n",
    "\n",
    "-   **Race Conditions:** In multithreaded environments, multiple threads might\n",
    "    attempt to create an instance simultaneously.\n",
    "-   **Double-Checked Locking:** Ensures that only one thread can create the\n",
    "    instance, preventing multiple instances from being created.\n",
    "\n",
    "## References And Further Readings\n",
    "\n",
    "-   [Singleton Design Pattern](https://refactoring.guru/design-patterns/singleton)\n",
    "-   [Making Singletons in Python](https://www.pythonmorsels.com/making-singletons/)"
   ]
  }
 ],
 "metadata": {
  "jupytext": {
   "cell_metadata_filter": "-all",
   "formats": "md:myst",
   "text_representation": {
    "extension": ".md",
    "format_name": "myst",
    "format_version": 0.13,
    "jupytext_version": "1.11.5"
   }
  },
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "mystnb": {
   "number_source_lines": true
  },
  "source_map": [
   16
  ]
 },
 "nbformat": 4,
 "nbformat_minor": 5
}